local P = {
	frame = CreateFrame("Frame","BG_Control-Core"),
	internalIDCount = 0
	}

local DB = {}
-- DB.nodeRadius = float
-- DB.nodesByID[...]
--			.position={"x"=x,"y"=y}
-- DB.playersByGUID[guid]
--			.class=CLASSNAME
--			.position={"x"=x,"y"=y}
--			.yardsFromPos=yards
--			.flags=bitFlags
--
--			.lastUpdate=time
--			.target=guid
--			.targedByMelee=guid
--
--			.spellID=(spellid or nil)
--			.raidID=(unitid or nil)
--			.spec=(specID or nil)
--			.name=targable name


function P.OnEvent(self, event, ...)
	-- Fires more than one a second
	if "COMBAT_LOG_EVENT_UNFILTERED" == event then
		P.Combat_Log_Event_Unfiltered(...);

	-- These are filtered combat logs, so they happen less frequently
	elseif "CHAT_MSG_ADDON" == event then
		--P.Chat_Msg_Addon(...);

	-- These updates happen semi-often on kills/point changes
	elseif "UPDATE_BATTLEFIELD_SCORE" == event then
		P.Update_Battlefield_Score(...);

	-- This fires only when someone leaves,joins, or groups move; hopefully not often
	elseif "GROUP_ROSTER_UPDATE" == event then
		P.Group_Roster_Update(...);

	-- These fire once a BG sessions
	elseif "PLAYER_ENTERING_BATTLEGROUND" == event then
		P.Player_Entering_Battleground(...);
	elseif "PLAYER_LEAVING_WORLD" == event then
		P.Player_Leaving_World(...);

	-- These things only happen "once" ever and then are unregistered
	elseif "WORLD_MAP_UPDATE" == event then
		P.World_Map_Update(...);
	elseif "ADDON_LOADED" == event then
		P.Addon_Loaded(...);
	end
end



P.frame:SetScript("OnEvent",P.OnEvent);
P.frame:RegisterEvent("ADDON_LOADED");

--====================================================================================
-- BEGIN


function P.Combat_Log_Event_Unfiltered(...)
	local timeStamp, event, _, sourceGUID, _, sourceFlags, _, destGUID, _, destFlags, _, spellID = ...
	P.ParseCombatLog(nil,event,timeStamp,sourceGUID,sourceFlags,destGUID,destFlags,spellID)
end


function P.ParseCombatLog(isForign,event,timeStamp,sourceGUID,sourceFlags,destGUID,destFlags,spellID)
	-- If this isn't both players, we can do NOTHING with it....
	if BG_ControlCommon.IsPlayer(sourceFlags) and BG_ControlCommon.IsPlayer(destFlags) then

		if "SWING_DAMAGE" == event then
			P.DamageDone(event,timeStamp,sourceGUID,sourceFlags,destGUID,destFlags, isForign, true)
		elseif "SPELL_CAST_SUCCESS" == event then
			P.SpellCastSuccess(event,timeStamp,destGUID,destFlags,spellID)
		elseif "RANGE_DAMAGE" == event then
			--local spellID= select(8,...)
			P.DamageDone(event,timeStamp,sourceGUID,sourceFlags,destGUID,destFlags,spellID, isForign)
		elseif "SPELL_DAMAGE" == event then
			--local spellID= select(8,...)
			P.DamageDone(event,timeStamp,sourceGUID,sourceFlags,destGUID,destFlags,spellID, isForign)
		elseif "SPELL_HEAL" == event then
			--local spellID= select(8,...)
			P.HealCast(event,timeStamp,sourceGUID,sourceFlags,destGUID,destFlags,spellID, isForign)
		elseif "UNIT_DIED" == event then
			P.PlayerDied(event,timeStamp,diedGUID,diedFlags)
		end
	end
end


function P.SpellCastSuccess(event,timeStamp,guid,flags,spellID)
	if BG_ControlCommon.IsHostile(flags) then
		-- If it's a "Mass Dispell"-able thing... go for it
		if spellID == 45438 or spellID == 642 or spellID == 1022 then
			-- If we can't find this player, quit
			if not DB.playersByGUID[guid] then return; end
			
			local pDB = DB.playersByGUID[guid];
			
			-- If it's old... we can't get a good "melee target"
			if not pDB.lastUpdate then return; end
			
			-- We don't know where they are anyway... quit
			if not pDB.position then return; end
			
			local internalNodeID = P.ClosestNodeTo( pDB.position)
			
			-- Bad node find, DOH
			if not internalNodeID then return; end
			
			-- None close, bail
			if not pDB.targedByMelee then return; end
			
			-- Can't tell them who... gtfo
			if not pDB.raidID then return; end
			
			BG_ControlInterface_EnemyInvulnerability(guid,spellID,internalNodeID,pDB.targedByMelee,pDB.raidID,pDB.lastUpdate,timeStamp)
		end
	end
end
function P.DamageDone(event,timeStamp,sourceGUID,sourceFlags,destGUID,destFlags,spellID, isForign, isMelee)
	-- TODO: Truth table this not insane like
	local friendlyOnTarget, friendlyGUID,enemyGUID, friendlyFlags, enemyFlags

	if BG_ControlCommon.IsFriendly(sourceFlags) and BG_ControlCommon.IsHostile(destFlags) then
		friendlyGUID = sourceGUID;
		enemyGUID = destGUID;
		friendlyFlags = sourceFlags;
		enemmyFlags = destFlags;
		friendlyOnTarget = true;
	elseif BG_ControlCommon.IsHostile(sourceFlags) and BG_ControlCommon.IsFriendly(destFlags)  then
		friendlyGUID = destGUID;
		enemyGUID = sourceGUID;
		friendlyFlags = destFlags;
		enemyFlags = sourceFlags;
	else
		-- THIS IS BROKEN, GTFO!
		return;
	end


	if not DB.playersByGUID[enemyGUID] then DB.playersByGUID[enemyGUID] = {}; end
	-- If the friendly doesn't have a friendly player... it's jacked up.  We'll get the next swing
	if not DB.playersByGUID[friendlyGUID] then BG_ControlCommon.Print("FAIL"..tostring(friendlyGUID)); P.Group_Roster_Update(); return; end

	-- This is an update (used when doing Addon cross-talk)
	local enemyDB = DB.playersByGUID[enemyGUID];
	
	if not enemyDB.lastUpdate or enemyDB.lastUpdate < timeStamp then
		local friendlyDB = DB.playersByGUID[friendlyGUID];

		-- No raidID = jacked up friendly... GTFO!
		if not friendlyDB.raidID then Group_Roster_Update(); return; end

		local x, y = GetPlayerMapPosition(friendlyDB.raidID)

		-- If it's invalid, get out!
		if not x or 0 == x or not y or 0 == y then return; end

		enemyDB.position = { ["x"]=x , ["y"]=y }

		local range
		if isMelee then
			range = 0
		else
			local _, _, _, _, _, maxRange = GetSpellInfo(spellID)
			range = maxRange
			P.TestSpec(enemyGUID,spellID,timeStamp)
		end
		enemyDB.yardsFromPos = range

		enemyDB.flags = enemyFlags

		enemyDB.lastUpdate = timeStamp
		enemyDB.target = toGUID

		P.UpdatePlayersNode(enemyGUID,timeStamp)

		friendlyDB.flags = friendlyFlags
		
		if friendlyOnTarget then
			if isMelee then
				BG_ControlInterface_FriendlyPlayerTarget(friendlyGUID,enemyGUID,friendlyDB.raidID);
			end
		else
			BG_ControlInterface_EnemyPlayerTarget(enemyGUID,friendlyGUID)
		end
		
	end

	if not isForign then
		-- TODO: Addon Cross Talk
		--P.SendCombatMessage(event,timeStamp,sourceGUID,sourceFlags,destGUID,destFlags)
	end
end

function P.TestSpec(guid,spellID,timeStamp)
	if not DB.playersByGUID[guid] then return; end
	
	-- We've already done this, cached, game over, ignore it...
	if DB.playersByGUID[guid].spec then return; end
	
	if not DB.playersByGUID[guid].class then
		local _, class, _, race, sex, name, realm = GetPlayerInfoByGUID(guid)
		local targetName = name
		if not realm or "" == realm then
			realm = GetRealmName()
		else
			targetName = targetName.."-"..realm
		end
		
		DB.playersByGUID[guid].class = class
		DB.playersByGUID[guid].name = targetName
		
		-- We're missing data we need! GET OUT!
		if class and race and sex and name then
			BG_ControlInterface_PlayerInformation(guid,targetName,class,name,realm,race,sex)
		end
	end
	local pDB = DB.playersByGUID[guid]
	
	local specID, class = BG_ControlCommon.SpecializationBySpell(spellID,pDB.class)
	
	if not pDB.class and class then pDB.class = class; end
	
	if specID then
		pDB.spec = specID;
		BG_ControlInterface_PlayerSpec(guid,specID,timeStamp)
		local _, specName= GetSpecializationInfoByID(specID)
		BG_ControlCommon.Print("Spec for: "..pDB.name.." is "..specName);
	end

end

function P.UpdatePlayersNode(guid)
	if not DB.playersByGUID[guid] then return; end
	local pDB = DB.playersByGUID[guid]

	if not pDB.position then return; end

	-- This is in error... need an update to be valid
	if not pDB.lastUpdate then return; end
	
	-- If this data is 15 seconds old... it's stale, tell them
	if pDB.lastUpdate + 10 < time() then
		BG_ControlInterface_EnemyStale(guid)
	else
		-- Or just go with the flow

		local internalNodeID,range,isClose = P.ClosestNodeTo( pDB.position)

		-- Jacked up position, GTFO
		if not internalNodeID then return; end

		local f = BG_ControlInterface_EnemyDamageAt;
		if pDB.spec then
			if BG_ControlCommon.IsHealer(pDB.spec) then			
				f = BG_ControlInterface_EnemyHealerAt;
			end
			
		end
		
		f(guid,internalNodeID,isClose,range,pDB.lastUpdate);
		
	end
end


function P.HealCast(event,timeStamp,sourceGUID,sourceFlags,destGUID,destFlags,spellID, isForign)
	-- We don't care about actions for our players... only enemy
	--	Also, self-healing gives us NO data... ignore it
	if BG_ControlCommon.IsHostile(sourceFlags) and
		BG_ControlCommon.IsHostile(destFlags) and
		sourceGUID ~= destGUID then

		if not DB.playersByGUID[sourceGUID] then DB.playersByGUID[sourceGUID] = {} end
		local healerDB = DB.playersByGUID[sourceGUID]

		-- If this person CAN be a healer, and they aren't flagged as one
		--	check the spell for a heal!
		-- TODO: When we know what talent spec they are and don't worry if they aren't a healer

		-- This is an update (used when doing Addon cross-talk)
		if not healerDB.lastUpdate or healerDB.lastUpdate < timeStamp then


			healerDB.flags=sourceFlags

			healerDB.lastUpdate = timeStamp
			-- TODO: Make this a function

			healerDB.target=destGUID;

			local targetDB = DB.playersByGUID[destGUID];
						-- Yup, we don't know about who they healed, bail
			if not targetDB then return; end

			-- We don't know where the target is located, so bail
			if not targetDB.position then return; end

			healerDB.position = targetDB.position

			local _, _, _, _, _, maxRange = GetSpellInfo(spellID)
			P.TestSpec(sourceGUID,spellID,timeStamp)
			healerDB.yardsFromPos = maxRange + ( targetDB.yardsFromPos or 0 )

			P.UpdatePlayersNode(sourceGUID,timeStamp)
		end

		-- If it's not from me... someone else got it too, so don't forward it
		if not isForign then
			--P.SendMessage(event, time,from,fromFlags,to,toFlags,spellID)
		end
	end
end

function P.PlayerDied(event,timeStamp,diedGUID,diedFlags)
	BG_ControlInterface_PlayerDied(diedGUID,diedFlags,timeStamp)
end

-- This is only "registered" if we have entered a PVP Battlegrounds.
function P.Player_Leaving_World()
	BG_ControlCommon.Print("Player has left PVP Battlegrounds");
	P.frame:UnregisterEvent("CHAT_MSG_ADDON");
	P.frame:UnregisterEvent("COMBAT_LOG_EVENT_UNFILTERED");
	P.frame:UnregisterEvent("PLAYER_LEAVING_WORLD");
	P.frame:UnregisterEvent("GROUP_ROSTER_UPDATE");
	-- TODO: This feature
	--P.frame:UnregisterEvent("UPDATE_BATTLEFIELD_SCORE");
	P.frame:UnregisterEvent("WORLD_MAP_UPDATE");

	-- TODO: Save off talent data to the saved variables
	BG_ControlInterface_BattlegroundEnded()
end


function P.Group_Roster_Update()
	-- If the BG isn't over
	if GetBattlefieldInstanceExpiration() == 0 then
		BG_ControlCommon.Print("Modifying group size...")
		for i=1,40 do
			local raidID = "raid"..i;
			local unitGUID = UnitGUID(raidID);
			-- If we haven't seen them "or we get it too soon" then sometimes this doesn't work
			if unitGUID then
				if not DB.playersByGUID[unitGUID] then DB.playersByGUID[unitGUID] = {}; end
				DB.playersByGUID[unitGUID].raidID=raidID;
			end

		end
	end
end

function P.Player_Entering_Battleground()
	local _, type, _, _, _, _, _, mapID = GetInstanceInfo()
	if "pvp" == type then
		BG_ControlCommon.Print("Player has entered PVP Battleground ID:".. mapID);

		P.InitDB();

		P.frame:RegisterEvent("CHAT_MSG_ADDON");
		P.frame:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED");
		P.frame:RegisterEvent("PLAYER_LEAVING_WORLD");
		P.frame:RegisterEvent("GROUP_ROSTER_UPDATE");
		-- TODO: This feature
		--BGC.frame:RegisterEvent("UPDATE_BATTLEFIELD_SCORE");
		P.frame:RegisterEvent("WORLD_MAP_UPDATE");
		BG_ControlInterface_BattlegroundStarted(mapID)

	end
end

function P.Addon_Loaded(...)
	local name = ...
	if "BG_Control" == name then
		BG_ControlCommon.Print("Loaded BG_Control!")
		P.frame:RegisterEvent("PLAYER_ENTERING_BATTLEGROUND");

		-- Get rid of this garbage... never used again
		P.frame:UnregisterEvent("ADDON_LOADED");
		P.Addon_Loaded = nil;
	end
end

function P.World_Map_Update()
	local oldSize = table.getn(P.nodesByID or {} );
	local landmarks = GetNumMapLandmarks();

	if landmarks == 0 then return end

	-- Map changed, we assume... get everything new!
	P.UpdateNodes()
	P.CalculateNodeRadius();
	
	
	for k,v in pairs(DB.playersByGUID) do
		P.UpdatePlayersNode(k)
	end
	
	BG_ControlInterface_NumbNodes(landmarks)

end

function P.UpdateNodes()
	local landmarks = GetNumMapLandmarks();
	
	-- Clear out the current node list, we're re-making it
	DB.nodesByID = {}
	for i=1,landmarks do
		local name, description, textureIndex, x, y, mapLinkID, inBattleMap, graveyardID, areaID, poiID, isObjectIcon, atlasIcon = GetMapLandmarkInfo(i);

		local internalID = P.GetInternalID(i,mapLinkID,graveyardID,areaID,poiID,isObjectIcon,atlasIcon,x,y);
		if not DB.nodesByID[internalID] then DB.nodesByID[internalID] = {}; end
		local node= DB.nodesByID[internalID]

		-- If it's new or has changed, we need to update it and recalculate



		-- If it's on the battle map, we care about it here too
		if inBattleMap then
			if not node.position then node.position = {}; end
			if  not node.position["x"] or not node.position["y"] or node.position["x"] ~= x or node.position["y"] ~= y then
				node.position = {["x"]=x,["y"]=y};
				needsNewRadius = true;
			end
			BG_ControlInterface_NodeUpdate(internalID,textureIndex,graveyardID,areaID,poiID,isObjectIcon,atlasIcon,name,description);
		end

	end
end

-- Let's just cover each node with a bubble that doesn't touch and say that is the "radius".
--	In reality, this should be converted and in yards, which Distance will take... eventaully...
function P.CalculateNodeRadius()
	local radius
	-- For each node, check the distance to every other node
	--	Yes, it is "doing extra work, eg) a->b and b-> a, but
	--	this is the price for readability
	for k,v in pairs(DB.nodesByID) do
		if v.position then
			local _,tmpRange = P.ClosestNodeTo(v.position);
			if not radius or tmpRange < radius then
				radius = tmpRange;
			end
		end
	end
	DB.nodeRadius = ( radius or 0 ) / 2;
end

function P.ClosestNodeTo(position)
	local node,range
	for k, v in pairs(DB.nodesByID) do
		-- TODO: Put in yards scaler
		local other = v.position
		-- Give an invalid range if they are the SAME location
		if other["x"] ~= position["x"] or other["y"] ~= position["y"] then
			local tmpRange = BG_ControlCommon.Distance(position,other);
			if tmpRange then
				if not node or tmpRange < range then
					range = tmpRange;
					node = k;
				end
			end
		end
	end
	-- If we can't find a node radius, then we can't say anything is near
	local isNear = ( range or 0 ) <= ( DB.nodeRadius or -100 );
	return node,range,isNear;
end

function P.InitDB()
	DB = {};
	DB.nodesByID = {};
	DB.playersByGUID = {};
	DB.nodeRadius = 0;
	P.ResetWorldMapPOI();
end

function P.GetInternalID(index,mapLinkID,graveyardID,areaID,poiID,isObjectIcon,atlasIcon,x,y)
	return index;
end

function P.ResetWorldMapPOI()
	for i=1,NUM_WORLDMAP_POIS do
		local f = _G["WorldMapFramePOI"..i]
		-- ERROR
		if not f then return; end
		f.bg_controlID = nil
	end
end